/*
optimbt.c Version 0.2.0  Btree optimal blocking factors
Copyright (C) 2003-2010  dondalah721@yahoo.com (Dondalah)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to:

	Free Software Foundation, Inc.
	59 Temple Place - Suite 330
	Boston, MA  02111-1307, USA.
*/

/* Usage: optimbt totrcds levels */

/* optimbt prints the optimum */
/* records per block */
/* based on number of levels in the btree */
/* Within reason, the fewer the levels in the btree */
/* the faster the access time */

#include <bthash.h>
#include <math.h>

void putstx(pgm)
char *pgm;
   {
   fprintf(stderr,"Usage: %s total_records "
      "levels\n",
      pgm);
   fprintf(stderr,"Where total_records is 1 - 999999999\n");
   fprintf(stderr,"      levels is 3 - 1023\n");
   exit(1);
   } /* putstx */

void bldrpt(totrcds,levels,nodesz)
int totrcds;
int levels;
int nodesz;
   {
   int lvl;
   int lvls;
   int lvltot[512];
   int lvlrem[512];
   int lvlblks[512];
   int blktot[512];
   btoptim(totrcds,nodesz,&lvls,
      lvltot, lvlrem, lvlblks, blktot);
   printf("Total Records     %9d\n", totrcds);
   printf("Records Per Block %9d\n\n", nodesz);
   printf("    Level   Records    Adjust   LvlBlks   TotBlks\n");
   lvl = 0;
   while (lvl < lvls)
      {
      printf("%9d %9d %9d %9d %9d\n",
	 lvl+1, lvltot[lvl], lvlrem[lvl],
	 lvlblks[lvl], blktot[lvl]);
      lvl++;
      } /* for each level */
   } /* bldrpt */

void calcleaf(totrcds,levels,nodesz)
int totrcds;
int levels;
int nodesz;
   {
   int lvls;
   int bestsz;
   int bestlvl;
   double ratio;
   double space;
   int lvltot[512];
   int lvlrem[512];
   int lvlblks[512];
   int blktot[512];
   btoptim(totrcds,nodesz-2,&lvls,
      lvltot, lvlrem, lvlblks, blktot);
   ratio = (double) lvltot[lvls-1]
      / (double) (nodesz-2);
   bestsz  = nodesz - 2;
   space   = ratio;
   bestlvl = lvls;
   printf("lvls  bestsz  bestlvl "
      "    ratio     space\n");
   printf("%4d %7d %8d   %f  %f\n",
      lvls, bestsz, bestlvl, ratio, space);
   btoptim(totrcds,nodesz,&lvls,
      lvltot, lvlrem, lvlblks, blktot);
   ratio = (double) lvltot[lvls-1]
      / (double) (nodesz);
   if (ratio > space)
      {
      bestsz = nodesz;
      space  = ratio;
      bestlvl = lvls;
      } /* better ratio */
   printf("%4d %7d %8d   %f  %f\n",
      lvls, bestsz, bestlvl, ratio, space);
   btoptim(totrcds,nodesz+2,&lvls,
      lvltot, lvlrem, lvlblks, blktot);
   ratio = (double) lvltot[lvls-1]
      / (double) (nodesz+2);
   if (ratio > space)
      {
      bestsz = nodesz+2;
      space  = ratio;
      bestlvl = lvls;
      } /* better ratio */
   printf("%4d %7d %8d   %f  %f\n",
      lvls, bestsz, bestlvl, ratio, space);
   putchar('\n');
   bldrpt(totrcds,bestlvl,bestsz);
   } /* calcleaf */

void calcall(totrcds)
int totrcds;
   {
   int nodesz;
   int levels;
   int bestsz;
   int bestlvl;
   double ratio;
   double space;
   int lvltot[512];
   int lvlrem[512];
   int lvlblks[512];
   int blktot[512];
   // printf("lvls  bestsz  bestlvl "
      // "    ratio\n");
   printf("levels    best nodesz  best level "
      "      ratio\n");
   space  = -999999.0;
   nodesz = 3;
   while (nodesz < 1024)
      {
      btoptim(totrcds,nodesz,&levels,
         lvltot, lvlrem, lvlblks, blktot);
      ratio = (double) lvltot[levels-1]
         / (double) nodesz;
      if (ratio > space)
         {
         bestsz = nodesz;
         space  = ratio;
         bestlvl = levels;
         printf("%6d %12d %11d       %f\n",
            levels, bestsz, bestlvl, ratio);
         // printf("%4d %7d %8d   %f\n",
            // levels, bestsz, bestlvl, ratio);
         } /* better ratio */
      nodesz += 2;
      } /* for each nodesize */
   putchar('\n');
   bldrpt(totrcds,bestlvl,bestsz);
   } /* calcall */

int main(argc,argv)
int argc;
char **argv;
   {
   int totrcds;
   int levels;
   int nodesz;
   double root;
   double dblrcds;
   double dbllvls;
   if (argc != 3) putstx(*argv);
   totrcds = atoi(*(argv+1));
   if (totrcds < 1 || totrcds > 999999999)
      {
      fprintf(stderr,"Total records is 1 - 999999999\n");
      putstx(*argv);
      } /* invalid totrcds */
   levels  = atoi(*(argv+2));
   if (levels < 1 || levels > 256)
      {
      fprintf(stderr,"Levels is 1 - 256\n");
      putstx(*argv);
      } /* invalid nodesz */

   dblrcds = (double) totrcds;
   dbllvls = (double) levels;
   root = pow(dblrcds, (1.0 / dbllvls));
   nodesz = (int) root + 1;
   if (!(nodesz & 1)) nodesz++;
   if (root < 3.0)
      {
      fprintf(stderr,"Too many levels for "
	 "total records\n");
      exit(1);
      } /* too many levels */
   if (root > 1023.0)
      {
      fprintf(stderr,"Not enough levels\n");
      fprintf(stderr,"Too many records "
	 "per node %12.1f\n",
	 root);
      exit(1);
      } /* too many levels */
   bldrpt(totrcds,levels,nodesz-2);
   putchar('\n');
   bldrpt(totrcds,levels,nodesz);
   putchar('\n');
   bldrpt(totrcds,levels,nodesz+2);
   putchar('\n');
   calcleaf(totrcds,levels,nodesz);
   putchar('\n');
   calcall(totrcds);
   return(0);
   } /* main */
